from dataclasses import dataclass

from dbt.adapters.snowflake.connections import (  # type:ignore
    SnowflakeConnectionManager,
    SnowflakeCredentials,
)

from common.credentials import LayerCredentials


@dataclass
class LayerSnowflakeCredentials(LayerCredentials, SnowflakeCredentials):
    @property
    def type(self) -> str:
        return "layer_snowflake"


class LayerSnowflakeConnectionManager(SnowflakeConnectionManager):
    TYPE = "layer_snowflake"



from dataclasses import dataclass
from typing import Any, Dict, Optional


@dataclass
class LayerCredentials:
    # Layer config json creds
    layer_configfile: Optional[str] = None
    layer_configfile_json: Optional[Dict[str, Any]] = None
    layer_project: Optional[str] = None
    layer_api_key: Optional[str] = None



from sklearn.metrics import accuracy_score, r2_score  # type: ignore
from sklearn.model_selection import GridSearchCV  # type: ignore
from xgboost import XGBClassifier, XGBRegressor

import layer

from .base_model import AutoMLModel, TrainDataset


class XGBoostClassifier(AutoMLModel):
    name = "XGBoost Classifier"
    model_type = AutoMLModel.CLASSIFIER

    def __init__(self) -> None:
        super().__init__()

    def train(self, ds: TrainDataset) -> None:
        hyperparameter_grid = {
            "max_depth": [2, 6, 10],
            "n_estimators": [60, 200],
            "learning_rate": [0.1, 0.01],
        }

        layer.log({"xgboost hyperparameter grid": hyperparameter_grid})

        estimator = XGBClassifier(seed=42, eval_metric="mlogloss", use_label_encoder=False)
        self.model: GridSearchCV = GridSearchCV(
            estimator=estimator, param_grid=hyperparameter_grid, scoring="accuracy", cv=2, n_jobs=-1, verbose=False
        )
        self.model.fit(ds.x_train, ds.y_train)

        layer.log({"xgboost best parameters": self.model.best_params_})
        preds = self.model.predict(ds.x_test)
        self.score = accuracy_score(ds.y_test, preds)
        self.feature_importances = self.model.best_estimator_.feature_importances_

    def compare_score(self, score: float) -> bool:
        return self.score > score


class XGBoostRegressor(AutoMLModel):
    name = "XGBoost XGBRegressor"
    model_type = AutoMLModel.REGRESSOR

    def __init__(self) -> None:
        super().__init__()

    def train(self, ds: TrainDataset) -> None:
        hyperparameter_grid = {
            "max_depth": [2, 6, 10],
            "n_estimators": [60, 200],
            "learning_rate": [0.1, 0.01],
        }

        layer.log({"xgboost hyperparameter grid": hyperparameter_grid})

        estimator = XGBRegressor(n_estimators=1000, objective="reg:squarederror")
        self.model: GridSearchCV = GridSearchCV(
            estimator=estimator, param_grid=hyperparameter_grid, cv=2, n_jobs=-1, verbose=False
        )
        self.model.fit(ds.x_train, ds.y_train)

        layer.log({"xgboost best parameters": self.model.best_params_})
        preds = self.model.predict(ds.x_test)
        self.score = r2_score(ds.y_test, preds)
        self.feature_importances = self.model.best_estimator_.feature_importances_

    def compare_score(self, score: float) -> bool:
        return self.score > score



from dbt.adapters.snowflake.impl import SnowflakeAdapter  # type:ignore

from common.adapter import LayerAdapter
from dbt.adapters.layer_snowflake.connections import LayerSnowflakeConnectionManager


class LayerSnowflakeAdapter(LayerAdapter, SnowflakeAdapter):
    ConnectionManager = LayerSnowflakeConnectionManager
    CASE_SENSITIVE = False



from typing import Any, Callable, List, Optional, Tuple

import sqlparse  # type:ignore
from sqlparse.sql import Token  # type:ignore
from sqlparse.tokens import Name, Newline, Punctuation  # type:ignore
from sqlparse.utils import remove_quotes  # type:ignore


class LayerSqlFunction:
    """
    A parsed Layer SQL statement
    """

    SUPPORTED_FUNCTION_TRAIN = "train"
    SUPPORTED_FUNCTION_PREDICT = "predict"
    SUPPORTED_FUNCTION_AUTOML = "automl"
    SUPPORTED_FUNCTION_TYPES = [SUPPORTED_FUNCTION_TRAIN, SUPPORTED_FUNCTION_PREDICT, SUPPORTED_FUNCTION_AUTOML]

    def __init__(self, function_type: str, source_name: str, target_name: str) -> None:
        self.function_type = function_type
        self.source_name = source_name
        self.target_name = target_name

    def __repr__(self) -> str:
        return (
            f"<LayerSQL function_type:{self.function_type}"
            + f" source_name:{self.source_name}"
            + f" target_name:{self.target_name}>"
        )


class LayerPredictFunction(LayerSqlFunction):
    def __init__(
        self,
        source_name: str,
        target_name: str,
        model_name: str,
        select_columns: List[str],
        predict_columns: List[str],
        all_columns: List[str],
        prediction_alias: str,
        sql: str,
    ) -> None:
        super().__init__(
            function_type=self.SUPPORTED_FUNCTION_PREDICT, source_name=source_name, target_name=target_name
        )
        self.model_name = model_name
        self.select_columns = select_columns
        self.predict_columns = predict_columns
        self.all_columns = all_columns
        self.prediction_alias = prediction_alias
        self.sql = sql


class LayerTrainFunction(LayerSqlFunction):
    def __init__(
        self,
        source_name: str,
        target_name: str,
        train_columns: List[str],
    ) -> None:
        super().__init__(function_type=self.SUPPORTED_FUNCTION_TRAIN, source_name=source_name, target_name=target_name)
        self.train_columns = train_columns


class LayerAutoMLFunction(LayerSqlFunction):
    def __init__(
        self,
        source_name: str,
        target_name: str,
        model_type: str,
        feature_columns: List[str],
        target_column: str,
        sql: str,
    ) -> None:
        super().__init__(function_type=self.SUPPORTED_FUNCTION_AUTOML, source_name=source_name, target_name=target_name)
        self.feature_columns = feature_columns
        self.target_column = target_column
        self.model_type = model_type
        self.sql = sql


def find_from_token(tokens: List[Token]) -> List[Token]:
    return expect_tokens(tokens, [keyword("from"), lambda x: isinstance(x, sqlparse.sql.Identifier)])


def get_from_where_clause(select_tokens: List[Token]) -> Tuple[str, str]:
    from_token = find_from_token(select_tokens)
    if not from_token:
        raise ValueError("Invalid sql. Missing 'from' clause.")
    source = from_token[0].value
    where_statement = " ".join((x.value for x in from_token[1:]))
    return source, where_statement


def get_cols_from_container(brace_container_token: Token) -> List[str]:
    _, cols_token, _ = brace_container_token.tokens
    if isinstance(cols_token, sqlparse.sql.IdentifierList):
        return [x.value for x in clean_separators(cols_token.tokens)]
    return [cols_token.value]


def build_sql(cols: List[str], source: str, where_statement: str) -> str:
    col_str = ", ".join(cols)
    if len(where_statement) > 0:
        return f"select {col_str} from {source} {where_statement}"
    else:
        return f"select {col_str} from {source}"


class LayerSQLParser:

    ALL_COLUMNS_WILDCARD = "*"

    def parse(self, sql: str) -> Optional[LayerSqlFunction]:
        """
        returns None if not a layer SQL statement
        returns an instance of LayerSQL if a valid layer SQL statement
        """
        parsed = sqlparse.parse(sql)
        if not parsed:
            return None

        tokens = parsed[0].tokens
        layer_func = next((x for x in find_functions(tokens) if is_layer_function(x)), None)
        if not layer_func:
            return None

        target_name_group = next(
            (x for x in expect_tokens(tokens, [keyword("create or replace"), keyword("table"), group()])), None
        )
        if not target_name_group:
            return None
        target_name = self._get_target_name_from_group(target_name_group)

        if is_automl_function(layer_func):
            return self.parse_automl(layer_func, target_name)
        elif is_predict_function(layer_func):
            return self.parse_predict(layer_func, target_name)
        elif is_train_function(layer_func):
            return self.parse_train(layer_func, target_name)
        else:
            raise ValueError(f"Unsupported function: {layer_func.get_name()}")

    def _get_target_name_from_group(self, token: Token) -> str:
        clean_inner_tokens = remove_tokens(token.flatten(), newline())
        target_name_tokens = slice_between_tokens(clean_inner_tokens, name(), whitespace())
        if target_name_tokens:
            return "".join([x.value for x in target_name_tokens])
        else:
            raise ValueError("Invalid sql syntax for Layer function")

    def parse_predict(self, layer_func_token: Token, target: str) -> LayerPredictFunction:

        # We need the parent `select` statement that contains the function
        # to get access to the selected columns and the source relation
        select_sttmt = find_parent(layer_func_token, lambda x: isinstance(x, sqlparse.sql.Parenthesis))
        clean_select_sttmt = []
        if not select_sttmt:
            raise ValueError("SQL syntax error")
        else:
            clean_select_sttmt = clean_separators(select_sttmt.tokens)

        # sqlparse doesn't seem to parse correctly the inner contents of this parenthesis.
        # Here, we rebuild the sql and use parse it again to get the relevant tokens
        inner_sql_text = " ".join(x.value for x in clean_select_sttmt)
        inner_sql = sqlparse.parse(inner_sql_text)[0]
        clean_inner_sql = remove_tokens(inner_sql.tokens, lambda x: x.is_whitespace)

        source, where_statement = get_from_where_clause(clean_inner_sql)

        # extract the selected columns in the query
        columns_incl_layer = []
        select_columns_list = find_token(clean_inner_sql, lambda x: isinstance(x, sqlparse.sql.IdentifierList))
        if select_columns_list is None:
            raise ValueError("SQL syntax error. Select columns are missing.")
        else:
            columns_incl_layer = find_tokens(
                select_columns_list.tokens, lambda x: isinstance(x, sqlparse.sql.Identifier)
            )
        columns = remove_tokens(columns_incl_layer, lambda t: find_layer_function(t) is not None)
        select_columns = [t.value for t in columns]

        # extract the layer prediction function
        clean_func_tokens = clean_separators(layer_func_token[1].tokens)
        if len(clean_func_tokens) < 3:
            invalid_func = " ".join(t.value for t in clean_func_tokens)
            raise ValueError(f"Invalid predict function syntax {invalid_func}")
        model_name, _, bracket_container = clean_func_tokens
        predict_model = remove_quotes(model_name.value)
        predict_columns = get_cols_from_container(bracket_container)

        all_columns = select_columns + list(set(predict_columns) - set(select_columns))
        sql_text = build_sql(all_columns, source, where_statement)
        sql = sqlparse.format(sql_text, keyword_case="lower", strip_whitespace=True)

        prediction_alias = layer_func_token.parent.get_alias() or "prediction"

        return LayerPredictFunction(
            source,
            target,
            predict_model,
            select_columns,
            predict_columns,
            all_columns,
            prediction_alias,
            sql,
        )

    def parse_automl(self, layer_func_token: Token, target: str) -> LayerAutoMLFunction:

        # We need the parent `select` statement that contains the function
        # to get access to the selected columns and the source relation
        select_sttmt = find_parent(layer_func_token, lambda x: isinstance(x, sqlparse.sql.Parenthesis))
        clean_select_sttmt = []
        if not select_sttmt:
            raise ValueError("SQL syntax error")
        else:
            clean_select_sttmt = clean_separators(select_sttmt.tokens)

        # sqlparse doesn't seem to parse correctly the inner contents of this parenthesis.
        # Here, we rebuild the sql and use parse it again to get the relevant tokens
        inner_sql_text = " ".join(x.value for x in clean_select_sttmt)
        inner_sql = sqlparse.parse(inner_sql_text)[0]
        clean_inner_sql = remove_tokens(inner_sql.tokens, lambda x: x.is_whitespace)

        source, where_statement = get_from_where_clause(clean_inner_sql)

        # extract the layer prediction function
        clean_func_tokens = clean_separators(layer_func_token[1].tokens)
        if len(clean_func_tokens) < 4:
            invalid_func = " ".join(t.value for t in clean_func_tokens)
            raise ValueError(f"Invalid automl function syntax {invalid_func}")
        model_type_token, _, bracket_container, target_column_token = clean_func_tokens
        model_type = remove_quotes(model_type_token.value)
        feature_columns = get_cols_from_container(bracket_container)
        target_column = remove_quotes(target_column_token.value)
        all_columns = list(dict.fromkeys((feature_columns + [target_column])))
        sql_text = build_sql(all_columns, source, where_statement)
        sql = sqlparse.format(sql_text, keyword_case="lower", strip_whitespace=True)

        return LayerAutoMLFunction(source, target, model_type, feature_columns, target_column, sql)

    def parse_train(self, layer_func_token: Token, target: str) -> LayerTrainFunction:
        select = find_parent(layer_func_token, lambda x: isinstance(x, sqlparse.sql.Identifier))
        if not select:
            raise ValueError("SQL syntax error")
        else:
            clean_select = clean_separators(select)
            source, _ = get_from_where_clause(layer_func_token.parent.parent.tokens)
            _, train_func = clean_select
            if len(train_func.tokens) < 2:
                invalid_func = "".join(x.value for x in select.flatten())
                raise ValueError(f"Invalid train function syntax {invalid_func}")
            _, parenthesis_group = train_func.tokens
            cols = self.extract_columns(clean_separators(parenthesis_group.tokens))

            return LayerTrainFunction(source, target, cols)

    def extract_columns(self, content_tokens: List[Token]) -> List[str]:
        if len(content_tokens) < 1:
            return [self.ALL_COLUMNS_WILDCARD]
        if content_tokens[0].value.lower() == "array":
            if len(content_tokens) < 2:
                raise ValueError("Invalid train function syntax")
            identifiers = clean_separators(content_tokens[1].tokens)[0]
            return [x.value for x in identifiers.get_identifiers()]
        else:
            return [content_tokens[0].value]


# Functions to assist the SQL Parsing


TokenPredicate = Callable[[Token], bool]


def get_predict_function_name(layer_column_token: Token) -> Optional[str]:
    if layer_column_token.has_alias():
        return layer_column_token.get_name()
    return "predict"


def find_token(tokens: List[Token], pred: TokenPredicate) -> Optional[Token]:
    for token in tokens:
        if pred(token):
            return token
    return None


def find_tokens(tokens: List[Token], pred: TokenPredicate) -> List[Token]:
    return [x for x in tokens if pred(x)]


def remove_tokens(tokens: List[Token], pred: TokenPredicate) -> List[Token]:
    return [x for x in tokens if not pred(x)]


def clean_separators(tokens: List[Token]) -> List[Token]:
    return remove_tokens(tokens, lambda x: x.is_whitespace or x.ttype == Punctuation)


def expect_tokens(tokens: List[Token], predicates: List[TokenPredicate]) -> List[Token]:
    if not tokens or not predicates:
        return tokens
    pred = predicates[0]
    for i, token in enumerate(tokens):
        if pred(token):
            return expect_tokens(tokens[i:], predicates[1:])
    return []


def slice_between_tokens(tokens: List[Token], start: TokenPredicate, end: TokenPredicate) -> Optional[List[Token]]:
    if not tokens:
        return None
    start_index = -1
    end_index = -1
    start_seen = False
    for i, token in enumerate(tokens):
        if not start_seen and start(token):
            start_index = i
            start_seen = True
        if start_seen and end(token):
            end_index = i
            break
    if start_index >= 0 and end_index > 0:
        return tokens[start_index:end_index]
    else:
        return None


def find_functions(elem: Any) -> List[Token]:
    from itertools import chain

    funcs: List[List[Token]] = [[]]
    if isinstance(elem, List):
        funcs = [find_functions(x) for x in elem]
    elif isinstance(elem, sqlparse.sql.Statement):
        funcs = [find_functions(x) for x in elem.tokens]
    elif isinstance(elem, sqlparse.sql.IdentifierList):
        funcs = [find_functions(x) for x in elem.get_identifiers()]
    elif isinstance(elem, sqlparse.sql.Identifier):
        funcs = [find_functions(x) for x in elem.tokens]
    elif isinstance(elem, sqlparse.sql.Parenthesis):
        funcs = [find_functions(x) for x in elem.tokens]
    elif isinstance(elem, sqlparse.sql.Function):
        return [elem]
    return list(chain.from_iterable(funcs))


def find_parent(token: Token, pred: TokenPredicate) -> Optional[Token]:
    parent = token.parent
    if not parent:
        return None
    elif pred(parent):
        return parent
    else:
        return find_parent(parent, pred)


def is_layer_function(func: sqlparse.sql.Function) -> bool:
    parent = func.parent
    return isinstance(parent, sqlparse.sql.Identifier) and parent.tokens[0].value == "layer"


def find_layer_function(tokens: List[Token]) -> Optional[Token]:
    return next((x for x in find_functions(tokens) if is_layer_function(x)), None)


def is_predict_function(func: sqlparse.sql.Function) -> bool:
    return func[0].value.lower() == LayerSqlFunction.SUPPORTED_FUNCTION_PREDICT


def is_automl_function(func: sqlparse.sql.Function) -> bool:
    return func[0].value.lower() == LayerSqlFunction.SUPPORTED_FUNCTION_AUTOML


def is_train_function(func: sqlparse.sql.Function) -> bool:
    return func[0].value.lower() == LayerSqlFunction.SUPPORTED_FUNCTION_TRAIN


def keyword(value: str) -> TokenPredicate:
    def check_token(token: Token) -> bool:
        return token.is_keyword and token.value.lower() == value.lower()

    return check_token


def whitespace() -> Callable[[Token], bool]:
    return lambda x: x.is_whitespace


def name() -> Callable[[Token], bool]:
    return lambda x: x.ttype is Name


def newline() -> Callable[[Token], bool]:
    return lambda x: x.ttype is Newline


def group() -> Callable[[Token], bool]:
    return lambda x: x.is_group



import tempfile
from dataclasses import dataclass
from importlib.machinery import SourceFileLoader
from pathlib import Path, PurePosixPath
from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple

import agate  # type: ignore
import cloudpickle  # type: ignore
import pandas as pd  # type: ignore
from dbt.adapters.base.impl import BaseAdapter  # type: ignore
from dbt.adapters.base.relation import BaseRelation  # type: ignore
from dbt.adapters.protocol import AdapterConfig  # type: ignore
from dbt.clients.jinja import MacroGenerator  # type: ignore
from dbt.context.providers import generate_runtime_model_context  # type: ignore
from dbt.contracts.connection import AdapterResponse  # type: ignore
from dbt.contracts.graph.manifest import Manifest, ManifestNode  # type: ignore
from dbt.events import AdapterLogger  # type: ignore
from dbt.exceptions import RuntimeException  # type: ignore

import layer
from layer.decorators import model as model_decorator

from . import pandas_helper
from .sql_parser import (
    LayerAutoMLFunction,
    LayerPredictFunction,
    LayerSQLParser,
    LayerTrainFunction,
)


logger = AdapterLogger("Layer")


@dataclass
class LayerAdapterResponse(AdapterResponse):
    """
    Layer Adapter response
    """


@dataclass
class LayerMeta:
    """
    Layer meta
    """

    entrypoint: str = "handler.py"
    fabric: Optional[str] = None


class LayerAdapter(BaseAdapter):  # pylint: disable=abstract-method
    """
    Layer specific overrides
    """

    LayerSQLParser = LayerSQLParser
    CASE_SENSITIVE = True

    def __init__(self, config: AdapterConfig):
        super().__init__(config)
        self.sql_parser = self.LayerSQLParser()
        self._manifest_lazy: Optional[Manifest] = None

    @property
    def _manifest(self) -> Manifest:
        if self._manifest_lazy is None:
            return self.load_manifest()
        return self._manifest_lazy

    def load_manifest(self) -> Manifest:
        if self._manifest_lazy is None:
            # avoid a circular import
            from dbt.parser.manifest import ManifestLoader  # type: ignore

            manifest = ManifestLoader.get_full_manifest(self.config)
            self._manifest_lazy = manifest
        return self._manifest_lazy

    def _get_manifest_node_from_relation_name(self, name: str) -> Optional[Tuple[ManifestNode, BaseRelation]]:
        for node in self._manifest.nodes.values():
            for suffix in ["", "__dbt_tmp"]:
                node = node.replace(alias=node.alias + suffix)
                relation = self.Relation.create_from_node(self.config, node)
                if relation.render() == name:
                    return node, relation
        return None

    def execute(
        self, sql: str, auto_begin: bool = False, fetch: bool = False
    ) -> Tuple[LayerAdapterResponse, agate.Table]:
        """
        if the given `sql` represents a Layer build or train, run Layer
        otherwise, pass the `execute` call to the underlying class
        """
        layer_sql_function = self.sql_parser.parse(sql)
        if layer_sql_function is None:
            return super().execute(sql, auto_begin, fetch)

        source_node_relation = self._get_manifest_node_from_relation_name(layer_sql_function.source_name)
        target_node_relation = self._get_manifest_node_from_relation_name(layer_sql_function.target_name)

        if not source_node_relation:
            raise RuntimeException(f'Unable to find a source named "{layer_sql_function.source_name}" in sql "{sql}"')
        if not target_node_relation:
            raise RuntimeException(f'Unable to find a target named "{layer_sql_function.target_name}" in sql "{sql}"')

        source_node, source_relation = source_node_relation
        target_node, target_relation = target_node_relation

        if isinstance(layer_sql_function, LayerTrainFunction):
            return self._run_layer_train(layer_sql_function, source_node, source_relation, target_node, target_relation)
        elif isinstance(layer_sql_function, LayerPredictFunction):
            return self._run_layer_predict(
                layer_sql_function, source_node, source_relation, target_node, target_relation
            )
        elif isinstance(layer_sql_function, LayerAutoMLFunction):
            return self._run_layer_automl(layer_sql_function, source_node, target_node)
        else:
            raise RuntimeException(f'Unknown layer function "{layer_sql_function.function_type}"')

    def _run_layer_train(
        self,
        layer_sql_function: LayerTrainFunction,
        source_node: ManifestNode,
        source_relation: BaseRelation,
        target_node: ManifestNode,
        target_relation: BaseRelation,
    ) -> Tuple[LayerAdapterResponse, agate.Table]:
        """
        Train a machine learning model using the given python script and save it as a dbt model
        """
        # load entrypoint
        entrypoint_module = self._get_layer_entrypoint_module(target_node)

        # load source dataframe
        raw_input_df = self._fetch_dataframe(source_node, source_relation)
        if layer_sql_function.train_columns == ["*"]:
            input_df = raw_input_df
        else:
            input_df = raw_input_df[layer_sql_function.train_columns].reset_index(drop=True)
        logger.debug("Fetched input dataframe - {}", input_df.shape)

        # login to Layer
        self.login_layer()

        # build the dataframe
        project_name = self.get_project_name(target_node)
        logger.debug("Training model {}, in project {}", target_node.name, project_name)
        layer.init(project_name)

        def training_func() -> Any:
            return entrypoint_module.main(input_df)

        model_decorator(project_name)(training_func)()  # pylint: disable=no-value-for-parameter
        logger.debug("Trained model {}, in project {}", target_node.name, project_name)

        output_df = pd.DataFrame.from_records([[target_node.name]], columns=["name"])

        # save the resulting dataframe to the target
        _, table = self._load_dataframe(target_node, output_df)

        response = LayerAdapterResponse(
            _message=f"LAYER MODEL TRAIN {output_df.shape[0]}",
            rows_affected=output_df.shape[0],
            code="LAYER TRAIN",
        )
        return response, table

    def login_layer(self) -> None:
        layer_api_key = self.config.credentials.layer_api_key
        if layer_api_key is not None:
            layer.login_with_api_key(layer_api_key)
        else:
            raise RuntimeException("Missing credentials: Please configure 'layer_api_key' in your 'profiles.yaml'.")

    def get_project_name(self, node: ManifestNode) -> str:
        if self.config.credentials.layer_project is not None:
            return self.config.credentials.layer_project
        return node.fqn[0]

    @staticmethod
    def _get_layer_meta(node: ManifestNode) -> LayerMeta:
        return LayerMeta(**node.meta.get("layer", {}))

    def _run_layer_automl(
        self,
        param: LayerAutoMLFunction,
        source_node: ManifestNode,
        target_node: ManifestNode,
    ) -> Tuple[LayerAdapterResponse, agate.Table]:
        input_df = self._fetch_dataframe_by_sql(source_node, param.sql)

        model_name = target_node.fqn[-1]

        # login to Layer
        self.login_layer()

        # init project
        project_name = self.get_project_name(target_node)
        logger.debug("Training AutoML model {}, in Layer project {}", model_name, project_name)
        layer.init(project_name)

        from .automl import AutoML

        automl = AutoML(param.model_type, input_df, param.feature_columns, param.target_column)
        automl.train(project_name, model_name)

        response = LayerAdapterResponse(
            _message="LAYER AUTOML COMPLETE",
            rows_affected=0,
            code="LAYER AUTOML",
        )

        output_df = pd.DataFrame(data=[[project_name, model_name]], columns=["project_name", "model_name"])
        _, table = self._load_dataframe(target_node, output_df)

        return response, table

    def _run_layer_predict(
        self,
        layer_sql_function: LayerPredictFunction,
        source_node: ManifestNode,
        source_relation: BaseRelation,
        target_node: ManifestNode,
        target_relation: BaseRelation,
    ) -> Tuple[LayerAdapterResponse, agate.Table]:
        try:
            # load source dataframe
            input_df = self._fetch_dataframe_by_sql(source_node, layer_sql_function.sql, layer_sql_function.all_columns)
            logger.debug("Fetched input dataframe - {}", input_df.shape)

            # Users can use the full path for fetching models. If they are authenticated, they can also use the model
            # name as the path. So, we try to log in and init project, only if we have the Layer api key in the dbt
            # profile
            if self.config.credentials.layer_api_key:
                self.login_layer()
                layer.init(self.get_project_name(target_node))

            # Fetch the model
            layer_model_def = layer.get_model(layer_sql_function.model_name)

            model_input = input_df[layer_sql_function.predict_columns]
            predictions = layer_model_def.predict(model_input)
            logger.debug("Prediction dataframe - {}", predictions.shape)
            column_template = layer_sql_function.prediction_alias
            prediction_column_count = len(predictions.columns)
            if prediction_column_count > 1:
                column_template += "_{ix}"
            predictions.columns = [column_template.format(ix=ix) for ix in range(prediction_column_count)]
            select_columns_from_source = list(set(layer_sql_function.select_columns) - set(predictions.columns))
            result_df = pd.concat(
                [
                    input_df[select_columns_from_source].reset_index(drop=True),
                    predictions.reset_index(drop=True),
                ],
                axis=1,
            )

            # save the resulting dataframe to the target
            _, table = self._load_dataframe(target_node, result_df)

            response = LayerAdapterResponse(
                _message=f"LAYER PREDICTION INSERT {predictions.shape[0]}",
                rows_affected=predictions.shape[0],
                code="LAYER PREDICT",
            )
            return response, table
        except Exception as e:
            import traceback

            traceback.print_exc()
            raise e

    def _get_layer_entrypoint_module(self, node: ManifestNode) -> ModuleType:
        """
        get the entrypoint absolute path
        - if entry point is not absolute, append it to the patch_path directory
        - if entry point is absolute, take it from the project root

        then load the module at that path
        """

        layer_meta = LayerMeta(**node.meta.get("layer", {}))

        entrypoint = PurePosixPath(layer_meta.entrypoint)

        if entrypoint.is_absolute():
            entrypoint = entrypoint.relative_to(entrypoint.root)
        else:
            _, patch_file_path = node.patch_path.split("://")
            entrypoint = PurePosixPath(patch_file_path).parent / entrypoint

        entrypoint = PurePosixPath(node.root_path) / entrypoint
        logger.debug("Loading Layer entrypoint at {}", entrypoint)

        entrypoint_module = SourceFileLoader(  # pylint: disable=deprecated-method
            f"layer_entrypoint.{node.unique_id}", str(entrypoint)
        ).load_module(None)

        # register this module to be pickled, otherwise pickling fails on dynamically created modules
        cloudpickle.register_pickle_by_value(entrypoint_module)

        return entrypoint_module

    def _fetch_dataframe(self, node: ManifestNode, relation: BaseRelation) -> pd.DataFrame:
        """
        Fetches all the data from the given node/relation and returns it as a pandas dataframe
        """
        # TODO: fix Possible SQL injection vector through string-based query construction. and remove nosec
        sql = f"select * from {relation.render()}"  # nosec
        return self._fetch_dataframe_by_sql(node=node, sql=sql)

    def _fetch_dataframe_by_sql(
        self, node: ManifestNode, sql: str, query_column_names: Optional[List[str]] = None
    ) -> pd.DataFrame:
        """
        Fetches all the data from the given sql and returns it as a pandas dataframe
        """
        # If case sensitive, don't map any columns
        # If not case sensitive, map upper columns to the names given in the query
        column_names_map = (
            {}
            if self.CASE_SENSITIVE or not query_column_names
            else {column.upper(): column for column in query_column_names}
        )

        with self.connection_for(node):
            # call super() instead of self to avoid a potential infinite loop
            unused_response, table = super().execute(sql, auto_begin=True, fetch=True)
            super().commit_if_has_connection()
            dataframe = pandas_helper.from_agate_table(table, column_names_map)

        return dataframe

    def _load_dataframe(self, node: ManifestNode, dataframe: pd.DataFrame) -> Tuple[Dict[Any, Any], agate.Table]:
        """
        Loads the given pandas dataframe into the given node/relation
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            file = Path(tmpdirname) / "data.csv"
            table = pandas_helper.to_agate_table_with_path(dataframe, file)

            materialization_macro = self._manifest.find_materialization_macro_by_name(
                self.config.project_name,
                "seed",
                self.type(),
            )

            context = generate_runtime_model_context(node, self.config, self._manifest)
            context["load_agate_table"] = lambda: table

            with self.connection_for(node):
                result = MacroGenerator(materialization_macro, context)()

        return result, table
